#
# Copyright (c) 2019, 2020 Oracle and/or its affiliates. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from fdk.async_http import protocol


def http_protocol(**kwargs):
    return protocol.HttpProtocol(
        loop=None, request_handler=None, error_handler=None,
        **kwargs)


def test_keep_alive_time_left_before_time_is_set():
    p = http_protocol()

    protocol.current_time = None
    p._last_request_time = None

    time_left = p.keep_alive_time_left()
    assert 0 == time_left


def test_keep_alive_time_left_after_time_is_set():
    p = http_protocol(keep_alive_timeout=15)

    protocol.current_time = 96
    p._last_response_time = 64

    time_left = p.keep_alive_time_left()
    assert 15 - 32 == time_left
